/* linspline subroutine library
 *     Michiel de Hoon (mdehoon@ims.u-tokyo.ac.jp)
 *     Laboratory of DNA Information Analysis, Human Genome Center
 *     Institute of Medical Science, University of Tokyo
 *     April 25, 2002
 *     Converted to ANSI-C on November 11, 2002.
 */

#include <stdlib.h>
#include <math.h>

/* ************************************************************************ */

double tserror (int nknots, int knots[], double b[], int n, double x[], int ny[], double sy[], double ssy[])
/*
 *  -- tserror routine (version 0) --
 *     Michiel de Hoon (mdehoon@ims.u-tokyo.ac.jp)
 *     Laboratory of DNA Information Analysis, Human Genome Center
 *     Institute of Medical Science, University of Tokyo
 *     April 17, 2002
 *
 *  Purpose
 *  =======
 *
 *  TSERROR computes the total squared error between a linear spline
 *  function and the data set to which the linear spline function was
 *  fitted.
 *
 *  Arguments
 *  =========
 *
 *  nknots  (input) int
 *          The number of knots of the linear spline function.
 *
 *  n       (input) int
 *          The number of data points in the data set.
 *
 *  knots   (input) int[nknots]
 *          The position of the knots, given in terms of the index into
 *          array x.
 *
 *  b       (input) double[nknots]
 *          The function value of the linear spline function at the knot
 *          positions.
 *
 *  x       (input) double[n]
 *          The x-coordinates of the data points.
 *
 *  ny      (input) int[n]
 *          The number of data points at each x-coordinate.
 *
 *  sy      (input) double[n]
 *          The sum of the y-coordinates of the datapoints
 *          at each x-coordinate.
 *
 *  ssy     (input) double[n]
 *          The squared sum of the y-coordinates of the data points
 *          at each x-coordinate.
 *
 *
 *  =====================================================================
 */
{ double result = 0.;
  int i = 0;
  /* Sum over all data points */
  int j;
  for (j = 0; j < n; j++)
  /* Find the knot interval for each data point */
  { 
    double term;
    while (j>knots[i]) i++;
    /* Find the linear spline interpolation for each data point */
    if (j==knots[i]) term = b[i];
    else
    { double frac = (x[j]-x[knots[i-1]])/(x[knots[i]]-x[knots[i-1]]);
      term =  (1. - frac) * b[i-1] + frac * b[i];
    }
    /* Find square error between the data point and the interpolated value */
    result += ssy[j] - 2. * sy[j] * term + ny[j] * term * term;
  }
  return result;
}

/* ************************************************************************ */

void fcombi (int ntotal, int n, int istart, int combi[], int* flag)
/*
 *
 *  -- fcombi routine (version 1) --
 *     Michiel de Hoon (mdehoon@ims.u-tokyo.ac.jp)
 *     Laboratory of DNA Information Analysis, Human Genome Center
 *     Institute of Medical Science, University of Tokyo
 *     April 17, 2002
 *
 *  Purpose
 *  =======
 *
 *  fcombi finds the next combination of n integers out of
 *  the range (istart,istart+1,,...,istart+ntotal-1).
 *
 *  Arguments
 *  =========
 *
 *  ntotal  (input) int
 *          The number of integers in the set from which a combination
 *          is chosen.
 *
 *  n       (input) int
 *          The number of integers that form a combination.
 *
 *  istart  (input) int
 *          The smallest integer in the set from which a combination is
 *          chosen.
 *
 *  combi   (input/output) int[n]
 *          The combination that is chosen.
 *          Upon entry, combi is the previous combination that
 *          was chosen.
 *          Upon exit, combi is the next combination to be used.
 *
 *  flag    (input/output) int*
 *          Upon entry: If flag == -1, then combi is initialized to the 
 *          first possible combination. Otherwise, the next combination
 *          is calculated.
 *          Upon exit: If the last possible combination is being returned,
 *          flag is set equal to zero. Otherwise, it is set to unity.
 *
 *  =====================================================================
 */
{ int i;
  int j = 1;
  int is, js;
  /* If flag==-1, initialize. */
  if (*flag==-1)
  { for (i = 0; i < n; i++) combi[i] = istart + i;
    if (ntotal==n || n==0) *flag = 0;
    else *flag = 1; /* This is the only possible set of knots. */
    return;
  }

  /* Otherwise, calculate next set of knots. */
  *flag = 0;
  while ( (combi[n-j]+j > istart+ntotal-1) && (j < n)) j++;
  is = combi[n-j];
  js = n-j-1;
  for (i = 1; i <= j; i++) combi[js+i] = is + i;

  *flag = 1;
  /* If there are no more knots after this one, set flag. */
  if (combi[0]==istart+ntotal-n) *flag = 0;
  return;
}

/* ************************************************************************ */

void initerror (int m, int n, double sy[], double ssy[], int lfixed, double* fixed, double* toterror, double* aic)
/*
 *  -- initerror routine (version 1) --
 *     Michiel de Hoon (mdehoon@ims.u-tokyo.ac.jp)
 *     Laboratory of DNA Information Analysis, Human Genome Center
 *     Institute of Medical Science, University of Tokyo
 *     April 17, 2002
 *
 *  Purpose
 *  =======
 *
 *  INITERROR calculates the total error which we find when fitting
 *  a flat line to the data. It also calculates the corresponding value
 *  of the AIC.
 *
 *  Arguments
 *  =========
 *
 *  m       (input) int
 *          The total number of data points.
 *
 *  n       (input) int
 *          The number of different time points.
 *
 *  y       (input) double[sum(ny)]
 *          The y-value of the data points.
 *
 *  sy      (input) double[n]
 *          The sum of the y-values of the data points at each
 *          time point.
 *
 *  ssy     (input) double[n]
 *          The sum of squares of the y-values of the data points
 *          at each time point.
 *
 *  lfixed  (input) int
 *          If LFIXED is non-zero, a fixed knot exists at time zero.
 *
 *  fixed   (input/output) double*
 *          On input, if a fixed knot exists at time zero, then fixed
 *          gives the function value of the linear spline function at
 *          the fixed knot.
 *          On output, if no fixed point exists at time zero, then
 *          fixed is set equal to the average of all data points. This
 *          corresponds to fitting a flat line to the data.
 *
 *  toterror (output) double*
 *          The total error resulting from fitting a flat line to the 
 *          data. The total error is defined as the sum of the squared
 *          error with respect to the flat line.
 *
 *  aic     (output) double*
 *          Akaike's Information Criterion, corresponding to a flat line
 *          fitted to the data. If there is no fixed point, then there is
 *          one free parameter. Otherwise, there are no free parameters.
 *          
 *  =====================================================================
 */
{ double term = 0.;
  int j;
  *toterror = 0.;
  for (j = 0; j < n; j++)
  { term += sy[j];
    *toterror += ssy[j];
  }

  if(lfixed) *toterror += (- 2.*(*fixed)*term + (*fixed)*(*fixed)*m);
  else
  { *fixed = term/m;
    *toterror -= term*(*fixed);
  }
  *aic = m * log(*toterror) + 2.;

  return;
}

/* ************************************************************************ */

void setmatrix (int n, double x[], int ny[], int nknots, int knots[], double d[], double e[], int lfixed)
/*
 *  -- setmatrix routine (version 0) --
 *     Michiel de Hoon (mdehoon@ims.u-tokyo.ac.jp)
 *     Laboratory of DNA Information Analysis, Human Genome Center
 *     Institute of Medical Science, University of Tokyo
 *     April 17, 2002
 *
 *  Purpose
 *  =======
 *
 *  setmatrix sets up the matrix on the left-hand-side of the equation
 *  to fit a linear spline function to the data. This matrix is
 *  essentially a weights matrix that depends on the time points of the
 *  measured data to the knots of the linear spline function. The matrix
 *  is tridiagonal and symmetric, so we only have to calculate the values
 *  on the diagonal and on the first subdiagonal.
 *
 *  Arguments
 *  =========
 *
 *  n       (input) int
 *          The number of different time points.
 *
 *  x       (input) double[n]
 *          The x-coordinates of the data points.
 *
 *  ny      (input) int[n]
 *          The number of data points at each x-coordinate.
 *
 *  nknots  (input) int
 *          The number of knots of the linear spline function.
 *
 *  knots   (input) int[nknots]
 *          The position of the knots, given in terms of the index into
 *          array x.
 *
 *  lfixed  (input) int
 *          If lfixed is non-zero, a fixed knot exists at time zero.
 *
 *  d       (output) double[nknots]
 *          The values of the matrix on its diagonal.
 *
 *  e       (output) double[nknots-1]
 *          The values of the matrix on its subdiagonal.
 *
 *  =====================================================================
 */
{ int js, jc, jf;
  int i, j;

  /* Reset the matrix to zero (d will be reset later) */
  for (j = 0; j < nknots-1; j++) e[j] = 0.;

  /* Calculate the first row of the matrix */
  jf = knots[1];
  if (lfixed)
  { for (j = 1; j < jf; j++)
    { double denom = x[jf]-x[0];
      double nom = (x[jf]-x[j]) * (x[j]-x[0]);
      denom *= denom;
      e[0] += ny[j] * nom/denom;
    }
    d[0] = 1.;
  }
  else
  { d[0] = ny[0];
    for (j = 1; j < jf; j++)
    { double denom = x[jf]-x[0];
      double nom = x[jf] - x[j];
      denom *= denom;
      d[0] += ny[j] * nom*nom/denom;
      nom *= (x[j]-x[0]);
      e[0] += ny[j] * nom/denom;
    }
  }

  /* Calculate the last row of the matrix */
  js = knots[nknots-2];
  d[nknots-1] = ny[n-1];
  for (j = js+1; j < n-1; j++)
  { double term = (x[j] - x[js]) / (x[n-1]-x[js]);
    d[nknots-1] += ny[j] * term * term;
  }

  /* Calculate the other rows */
  for (i = 1; i < nknots-1; i++)
  { js = knots[i-1];
    jc = knots[i];
    jf = knots[i+1];
    d[i] = ny[jc];
    for (j = js+1; j < jc; j++)
    { double term = (x[js]-x[j])/(x[jc]-x[js]);
      d[i] += ny[j] * term * term;
    }
    for (j = jc+1; j < jf; j++)
    { double denom = x[jf]-x[jc];
      double nom = x[jf]-x[j];
      denom *= denom;
      d[i] += ny[j] * nom*nom/denom;
      nom *=  (x[j]-x[jc]);
      e[i] += ny[j] * nom/denom;
    }
  }
  return;
}

/* ************************************************************************ */

void setvector (int n, double x[], double sy[], int nknots, int knots[], double b[], int lfixed, double fixed)
/*
 *  -- setvector routine (version 0) --
 *     Michiel de Hoon (mdehoon@ims.u-tokyo.ac.jp)
 *     Laboratory of DNA Information Analysis, Human Genome Center
 *     Institute of Medical Science, University of Tokyo
 *     April 17, 2002
 *
 *  Purpose
 *  =======
 *
 *  setvector sets up the vector on the right-hand-side of the equation
 *  to fit a linear spline function to the data. This vector represents
 *  the y-value of the data points at the locations of the knots of the
 *  linear spline function. The contribution of each data point is
 *  weighted by the distance of the data point to the knot.
 *
 *  Arguments
 *  =========
 *
 *  n       (input) int
 *          The number of different time points.
 *
 *  x       (input) double[n]
 *          The x-coordinates of the data points.
 *
 *  sy      (input) double[n]
 *          The sum of the y-values of the data points at each
 *          x-coordinate.
 *
 *  nknots  (input) int
 *          The number of knots of the linear spline function.
 *
 *  knots   (input) int[nknots]
 *          The position of the knots, given in terms of the index into
 *          array X.
 *
 *  b       (output) double[nknots]
 *          The values of the vector on the right hand side.
 *
 *  lfixed  (input) int
 *          If lfixed is non-zero, a fixed knot exists at time zero.
 *
 *  fixed   (input) double
 *          If a fixed knot exists at time zero, then fixed gives the
 *          function value of the linear spline function at the fixed
 *          knot.
 *
 *  =====================================================================
 */
{ int js, jc, jf;
  double denom;
  int i,j;

  /* First element */
  if (lfixed) b[0] = fixed;
  else
  { b[0] = sy[0];
    jf = knots[1];
    denom = x[jf]-x[0];
    for (j = 1; j <= jf; j++)
    { double term = (x[jf]-x[j])/denom;
      b[0] += term * sy[j];
    }
  }

  /* Last element */
  b[nknots-1] = sy[n-1];
  js = knots[nknots-2];
  denom = x[n-1]-x[js];
  for (j = js+1; j < n-1; j++)
  { double frac = (x[j]-x[js])/denom;
    b[nknots-1] += frac*sy[j];
  }

  /* Other elements: */

  for (i = 1; i < nknots - 1; i++)
  { js = knots[i-1];
    jc = knots[i];
    jf = knots[i+1];
    b[i] = sy[jc];
    denom = x[jc]-x[js];
    for (j = js+1; j < jc; j++)
    { double frac = (x[j]-x[js])/denom;
      b[i] += frac * sy[j];
    }
    denom = x[jf]-x[jc];
    for (j = jc + 1; j < jf; j++)
    { double frac = (x[jf]-x[j])/denom;
      b[i] += frac * sy[j];
    }
  }
  return;
}

/* ************************************************************************ */

void factor (int n, double d[], double e[])
/*
 *  -- factor routine (version 0) --
 *     Michiel de Hoon (mdehoon@ims.u-tokyo.ac.jp)
 *     Laboratory of DNA Information Analysis, Human Genome Center
 *     Institute of Medical Science, University of Tokyo
 *     April 17, 2002
 *
 *  -- Based on LAPACK routine DPTTRF (version 3.0) --
 *     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
 *     Courant Institute, Argonne National Lab, and Rice University
 *     June 30, 1999
 *
 *  Purpose
 *  =======
 *
 *  factor computes the L*D*L' factorization of a real symmetric
 *  positive definite tridiagonal matrix A.  The factorization may also
 *  be regarded as having the form A = U'*D*U.
 *
 *  Arguments
 *  =========
 *
 *  n       (input) int
 *          The order of the matrix A.  N >= 0.
 *
 *  d       (input/output) double[n]
 *          On entry, the n diagonal elements of the tridiagonal matrix
 *          A.  On exit, the n diagonal elements of the diagonal matrix
 *          D from the L*D*L' factorization of A.
 *
 *  e       (input/output) double[n-1]
 *          On entry, the (n-1) subdiagonal elements of the tridiagonal
 *          matrix A.  On exit, the (n-1) subdiagonal elements of the
 *          unit bidiagonal factor L from the L*D*L' factorization of A.
 *          E can also be regarded as the superdiagonal of the unit
 *          bidiagonal factor U from the U'*D*U factorization of A.
 *
 *  =====================================================================
 */

{ /* Compute the L*D*L' (or U'*D*U) factorization of A. */
  int i4 = (n-1) % 4;
  int i;
  double ei;

  for (i = 0; i < i4; i++)
  { ei = e[i];
    e[i] = ei / d[i];
    d[i+1] -= e[i] * ei;
  }
  for (i = i4; i < n-4; i += 4)
  { /* Solve for e(i) and d(i+1). */
    ei =e[i];
    e[i] = ei / d[i];
    d[i+1] -= e[i]*ei;
    /* Solve for e(i+1) and d(i+2). */
    ei = e[i+1];
    e[i+1] = ei / d[i+1];
    d[i+2] -= e[i+1]*ei;
    /* Solve for e(i+2) and d(i+3). */
    ei = e[i+2];
    e[i+2] = ei / d[i+2];
    d[i+3] -= e[i+2]*ei;
    /* Solve for e(i+3) and d(i+4). */
    ei = e[i+3];
    e[i+3] = ei / d[i+3];
    d[i+4] -= e[i+3]*ei;
  }
  return;
}

/* ************************************************************************ */

void solve (int n, double d[], double e[], double b[])
/*
 *  -- Based on LAPACK routine DPTSV (version 3.0) --
 *     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
 *     Courant Institute, Argonne National Lab, and Rice University
 *     June 30, 1999
 *
 *  -- solve routine (version 0) --
 *     Michiel de Hoon (mdehoon@ims.u-tokyo.ac.jp)
 *     Laboratory of DNA Information Analysis, Human Genome Center
 *     Institute of Medical Science, University of Tokyo
 *     April 17, 2002
 *
 *  Purpose
 *  =======
 *
 *  =====================================================================
 */
{ int i;
  /* Compute the L*D*L' (or U'*D*U) factorization of A. */

  /* Solve L * x = b. */
  for (i = 1; i < n; i++) b[i] -= b[i-1]*e[i-1];
  /* Solve D * L' * x = b. */
  b[n-1] = b[n-1] / d[n-1];
  for (i = n - 2; i >= 0; i--) b[i] = b[i] / d[i] - b[i+1] * e[i];
  return;
}

/* ************************************************************************ */

void icopy (int n, int source[], int dest[], int stride)
{ int i;
  for (i = 0; i < n; i++) dest[i*stride] = source[i*stride];
  return;
}

void dcopy (int n, double source[], double dest[], int stride)
{ int i;
  for (i = 0; i < n; i++) dest[i*stride] = source[i*stride];
  return;
}

/* ************************************************************************ */

void linspline (int n, double x[], int ny[], double y[], int* nknots, int knots[], double yknots[], int lfixed, double fixed, double* relerror, int* info)
/*
 *  -- linspline routine (version 1) --
 *     Michiel de Hoon (mdehoon@ims.u-tokyo.ac.jp)
 *     Laboratory of DNA Information Analysis, Human Genome Center
 *     Institute of Medical Science, University of Tokyo
 *     April 17, 2002
 *
 *     Based on LAPACK version 3.0
 *     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
 *     Courant Institute, Argonne National Lab, and Rice University
 *     February 25, 1997
 *
 *  Purpose
 *  =======
 *
 *  linspline computes a linear spline function to fit a given data set.
 *
 *  Arguments
 *  =========
 *
 *  n       (input) int
 *          The number of different x-coordinates of the
 *          data points in the data set.
 *
 *  x       (input) double[n]
 *          The set of x-coordinates occurring in the data set.
 *
 *  ny      (input) int[n]
 *          The number of data points at each x-coordinate.
 *
 *  y       (input) double [sum(NY)]
 *          The y-coordinates of the datapoints sorted by 
 *          their x-coordinate.
 *
 *  nknots  (output) int*
 *          The number of knots of the linear spline function.
 *
 *  knots   (output) int[n]
 *          The position of the knots, given in terms of the index into
 *          array x. Only the array arguments [0:nknots-1] are used.
 *
 *  yknots  (output) double[n]
 *          The function value of the linear spline function at the knot
 *          positions. Only the array arguments ( 0:NKNOTS-1 ) are used.
 *
 *  lfixed  (input) int
 *          If lfixed is non-zero, a fixed knot exists at time zero.
 *
 *  fixed   (input) double
 *          If a fixed knot exists at time zero, then fixed gives the
 *          function value of the linear spline function at the fixed
 *          knot.
 *
 *  error   (output) double
 *          computes the total squared error between a linear spline
 *          function and the data set to which the linear spline function
 *          was fitted, divided by the total square error if a flat line
 *          would be fitted to the data (i.e., the relative error). This
 *          is equal to 1.-(fraction variance explained).
 *
 *  info    (output) int*
 *          = 0:  successful exit
 *          > 0:  if info == 1, there was at least one time point at
 *                which there are no measurements. The linear spline
 *                function cannot be found. 
 *
 *  =====================================================================
 */
{
  int i;
  int m;
  int tnk;
  double term;
  double toterror = 0.;
  double aic = 0.;
  double* sy;
  double* ssy;
  int* tknots;
  *info = 0;
  *nknots = 0;

  /* First make sure we have data points where needed. */
  if (lfixed==0 && ny[0]==0)
  { *info = 1;
    return;
  }
  for (i = 1; i < n; i++)
  { if (ny[i]<0) 
    { *info = 1;
      return;
    }
  }

  /* Find the total number of data points, as well as the moments of Y.
   * Time point zero is treated as a special case.
   */
  sy = (double*)malloc((size_t)n*sizeof(double));
  ssy = (double*)malloc((size_t)n*sizeof(double));;
  if (lfixed==0)
  { sy[0] = 0.;
    ssy[0] = 0.;
    for (i = 0; i < ny[0]; i++)
    { term = y[i];
      sy[0] += term;
      ssy[0] += term*term;
    }
  }
  else
  { sy[0] = 0.;
    ssy[0] = 0.;
  }

  m = ny[0];
  for (i = 1; i < n; i++)
  { int j;
    sy[i] = 0.;
    ssy[i] = 0.;
    for (j = 0; j < ny[i]; j++)
    { term = y[j+m];
      sy[i] += term;
      ssy[i] += term*term;
    }
    m += ny[i];
  }

  /* Find the error for a flat line */
  initerror (m, n, sy, ssy , lfixed, &fixed, &toterror, &aic);
  yknots[0] =fixed;
  *nknots = 1;
  *relerror = 1.;

  /* Find the error for two or more knots */
  tknots = (int*)malloc((size_t)n*sizeof(int));
  tknots[0] = 0;
  for (tnk = 2; tnk <= n; tnk++)
  { int flag = -1;
    double* b = (double*)malloc((size_t)tnk*sizeof(double));
    double* d = (double*)malloc((size_t)tnk*sizeof(double));
    double* e = (double*)malloc((size_t)(tnk-1)*sizeof(double));
    tknots[tnk-1] = n-1;
    /* Find new knots and corresponding aic */
    while (flag)
    { double error;
      double taic;
      fcombi (n-2, tnk-2, 1, &tknots[1], &flag);
      setmatrix (n,x,ny,tnk, tknots, d, e, lfixed);
      setvector (n,x,sy,tnk, tknots, b, lfixed, fixed);
      if (lfixed)
      { b[1] = b[1] - e[0] * fixed;
        e[0] = 0.;
      }
      /* Compute the L*D*L' (or U'*D*U) factorization of A. */
      factor (tnk, d, e);
      /* Solve the system A*X = B, overwriting B with X. */
      solve (tnk, d, e, b);
      /* Calculate the error */
      error = tserror (tnk, tknots, b, n, x, ny, sy, ssy);
      /* Calculate the AIC (some constant terms are left out) */
      taic = m * log(error) + 2 * tnk;
      /* Compare this aic with the best found so far */
      if (taic < aic)
      { int stride = 1;
        *nknots = tnk;
        icopy (tnk,tknots,knots,stride);
        dcopy (tnk,b,yknots,stride);
        *relerror = error / toterror;
        aic = taic;
      }
    }
    free(b);
    free(d);
    free(e);
  }
  free(sy);
  free(ssy);
  free(tknots);
  return; 
}
