#ifdef __CPLUSPLUS__
extern "C" {
#endif
#include "Python.h"
#include "Numeric/arrayobject.h"
 
static PyObject *ErrorObject;

static int array_really_contiguous(PyArrayObject *ap) {
      int sd;
      int i;

      sd = ap->descr->elsize;
      for (i = ap->nd-1; i >= 0; --i) {
              if (ap->dimensions[i] == 0) return 1; /* contiguous by definition */
              if (ap->strides[i] != sd) return 0;
              sd *= ap->dimensions[i];
      }
      return 1;
}

struct fcomplex {
    float r;
    float i;
    };
typedef struct fcomplex Py_complex_float;
#define TRANSPOSE_OPTION 0
#define MIRROR_OPTION 0
#define get_fortran_dim(v,n) v->dimensions[(n)-1]

/* 
    Built by PyFort for C compiler.
*/

static int default_option = TRANSPOSE_OPTION;
static PyObject*
set_pyfort_option (PyObject* unused, PyObject* args) {
    if(!PyArg_ParseTuple(args, "i", &default_option)) return NULL;
    Py_INCREF(Py_None);
    return Py_None;
}

static void
set_pyfort_error (char* routine, char* var, char* problem) {
    char buf[512];
    sprintf(buf, "%s, argument %s: %s", routine, var, problem);
    PyErr_SetString (ErrorObject, buf);
}

static void set_transposed_strides (PyArrayObject* ar)
{
    int m1, n1, itmp; 
    n1 = ar->nd; 
    if (n1 < 2) return;
    m1 = ar->strides[n1-1];   /* stride for one element */ 
    for(itmp=0; itmp < n1 ; itmp++) { 
        ar->strides[itmp] = m1; 
        m1 *= ar->dimensions[itmp]; 
    } 
    ar->flags &= ~CONTIGUOUS; 
}


static PyArrayObject*
transpose_array (char* rname, char* vname, PyArrayObject* ap) {
/* return transpose of ap, possibly not contiguous so as to avoid copy if we
   are transposing a previously transposed contiguous array
   This means with the transpose option on the output of one call might
   not need any copying if used as input to another call. I.e., Fortran 
   arrays stay in row-major order.
*/
    int i, n;
    PyArrayObject *ret;
    n  = ap->nd;

    /* this allocates memory for dimensions and strides (but fills them
           incorrectly), sets up descr, and points data at ap->data. */
    ret = (PyArrayObject *)PyArray_FromDimsAndData(n, ap->dimensions,
                                                    ap->descr->type_num,
                                                    ap->data);
    if (!ret) {
        set_pyfort_error (rname, vname, "Could not create descriptors for transpose.");
        return NULL;
    }
       
    /* point at true owner of memory: */
    ret->base = (PyObject *)ap;
    Py_INCREF(ap);
    for(i=0; i<n; i++) {
        ret->dimensions[i] = ap->dimensions[n - i - 1];
        ret->strides[i] = ap->strides[n - i - 1];
    }
    if (array_really_contiguous(ret)) {
        ret->flags |= CONTIGUOUS;
    } else {
        ret->flags &= ~CONTIGUOUS;
    }
    return ret;
}

static PyArrayObject* make_contiguous(char* rname, char* vname, PyArrayObject* ap)
{
/* return an owned ref to a contiguous version of ap */
    PyArrayObject* result;
    if (ap->flags & CONTIGUOUS) {
        Py_INCREF (ap);
        return ap;
    } else {
        result = (PyArrayObject *) PyArray_ContiguousFromObject(
			(PyObject*) ap, ap->descr->type_num, 0, 0);
        if(!result) set_pyfort_error(rname, vname, "Failed making object contiguous.");
        return result;
    }
}

static int do_size_check (char* rname, char* vname, PyArrayObject *av, int rank,  int extents[], int mirror)
{
    int size1;
    int i, j;
    char buf[512];

    size1 = av->nd;
    
    if( size1 == rank) {
        for(i=0; i < rank; ++i) {
            /* no checking on last dimension of expected size 1 */
            if (i == size1-1) {
               if (extents[i] == 1) break;
            }
            j = mirror ? size1 - 1 - i : i;
            if(av->dimensions[j] != extents[i]) 
            {
               sprintf(buf, "Incorrect extent in dimension %d (%d expected %d)",
                       i+1, av->dimensions[j], extents[i]);
               set_pyfort_error(rname, vname, buf);
               return 0;
            }
        } 
    } else {
        if (rank != 1 || 
            size1 > 0 ||
            extents[0] != 1) 
        {    
           sprintf(buf, "Incorrect rank (%d expected %d)",
                       size1, rank);
           set_pyfort_error(rname, vname, buf);
           return 0;
        }
    }
    return 1; /* size ok */
}

static PyArrayObject*
do_array_in (char* rname, char* vname, PyObject *v, 
    enum PyArray_TYPES python_array_type)
{
    PyArrayObject* av;
    PyArrayObject* t;

    if(!PyArray_Check (v)) {
        t = (PyArrayObject *) PyArray_ContiguousFromObject(v, PyArray_NOTYPE, 0, 0);
        if (!t) {
            set_pyfort_error(rname, vname, "Argument cannot be converted to needed array.");
            goto err;
        }
    } else {
        t = (PyArrayObject*) v;
        Py_INCREF((PyObject*) t);
    }
    if (t->descr->type_num != python_array_type) {
        av = (PyArrayObject*) PyArray_Cast (t, python_array_type);
        Py_DECREF((PyObject*) t);
        t = av;
        if (!t) {
            set_pyfort_error(rname, vname, "Argument cannot be cast to needed type.");
            goto err;
        }
    } 
    return t;

err:
   return (PyArrayObject*) 0;
}

static PyArrayObject*
do_array_inout (char* rname, char* vname, PyObject *v, 
    enum PyArray_TYPES python_array_type)
{
    PyArrayObject* av;

   if (!PyArray_Check (v)) {
        set_pyfort_error(rname, vname, "Argument intent(inout) must be an array.");
        goto err;
   }
   av = (PyArrayObject*) v;
   if (av->descr->type_num != python_array_type) {
        set_pyfort_error(rname, vname, "Argument intent(inout) must be of correct type.");
        goto err;
   }
   if (!(av->flags & CONTIGUOUS))  {
       set_pyfort_error(rname, vname, "Argument intent(inout) must be contiguous.");
       goto err;
   }
   Py_INCREF(av);
   return av;
err:
   return (PyArrayObject*) 0;
}

static PyArrayObject*
do_array_create (char* rname, char* vname, enum PyArray_TYPES python_array_type, 
    int rank, int extents[], int mirror)
{
    PyArrayObject* av;
    int i, dims[7];
    if (rank > 7) {
        set_pyfort_error(rname, vname, "Too many dimensions -- limit is 7.");
        goto err;
    }
    if(mirror) {
        for(i=0; i < rank; ++i) dims[i] = extents[rank-1-i];
    } else {
        for(i=0; i < rank; ++i) dims[i] = extents[i];
    }
    av = (PyArrayObject*) PyArray_FromDims(rank, dims, python_array_type);
    if (!av) {
        set_pyfort_error(rname, vname, "Could not create array -- too big?");
        goto err;
    }
    return av;
err:
    return (PyArrayObject*) 0;
}
/* Methods */

/* linspline */
static char linspline_linspline__doc__[] =
"	X, NY, Y, LFIXED, FIXED\n"
"\n"
"\n"
" -- LINSPLINE routine (version 1) --\n"
"    Michiel de Hoon (mdehoon@ims.u-tokyo.ac.jp)\n"
"    Laboratory of DNA Information Analysis, Human Genome Center\n"
"    Institute of Medical Science, University of Tokyo\n"
"    February 4, 2002\n"
"\n"
"    Based on LAPACK version 3.0\n"
"    Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,\n"
"    Courant Institute, Argonne National Lab, and Rice University\n"
"    February 25, 1997\n"
"\n"
"\n"
" Purpose\n"
" =======\n"
"\n"
" LINSPLINE computes a linear spline function to fit a given data set.\n"
"\n"
" Arguments\n"
" =========\n"
"\n"
" N       (input) INTEGER\n"
"         The number of different x-coordinates of the\n"
"         data points in the data set.\n"
"\n"
" X       (input) DOUBLE PRECISION array, dimension ( 0:N-1 )\n"
"         The set of x-coordinates occurring in the data set.\n"
"\n"
" NY      (input) INTEGER array, dimension ( 0:N-1 )\n"
"         The number of data points at each x-coordinate.\n"
"\n"
" Y      (input) DOUBLE PRECISION array, dimension ( sum(NY) )\n"
"         The y-coordinates of the datapoints sorted by\n"
"         their x-coordinate.\n"
"\n"
" NKNOTS  (output) INTEGER\n"
"         The number of knots of the linear spline function.\n"
"\n"
" KNOTS   (output) INTEGER array, dimension ( 0:N-1)\n"
"         The position of the knots, given in terms of the index into\n"
"         array X. Only the array arguments ( 0:NKNOTS-1 ) are used.\n"
"\n"
" YKNOTS  (output) DOUBLE PRECISION array, dimension ( N-1 )\n"
"         The function value of the linear spline function at the knot\n"
"         positions. Only the array arguments ( 0:NKNOTS-1 ) are used.\n"
"\n"
" LFIXED  (input) INTEGER\n"
"         If LFIXED is non-zero, a fixed knot exists at time zero.\n"
"\n"
" FIXED   (input) DOUBLE PRECISION\n"
"         If a fixed knot exists at time zero, then FIXED gives the\n"
"         function value of the linear spline function at the fixed\n"
"         knot.\n"
"\n"
" ERROR   (output) DOUBLE PRECISION\n"
"         computes the total squared error between a linear spline\n"
"         function and the data set to which the linear spline function\n"
"         was fitted, divided by the total square error if a flat line\n"
"         would be fitted to the data (i.e., the relative error). This\n"
"         is equal to 1.-(fraction variance explained).\n"
"\n"
" INFO    (output) INTEGER\n"
"         = 0:  successful exit\n"
"         > 0:  if INFO = 1, there was at least one time point at\n"
"               which there are no measurements. The linear spline\n"
"               function cannot be found.\n"
"\n"
" =====================================================================\n"
"\n"
"    .. Scalar Arguments: INFO, N, LFIXED, FIXED\n"
"    .. Array Arguments:  X, NY, Y\n";

extern void linspline(long N, double* X, long* NY, double* Y, long* NKNOTS, long* KNOTS, double* YKNOTS, long LFIXED, double FIXED, double* RELERROR, long* INFO);

static PyObject*
linspline_linspline (PyObject* unused, PyObject* args) {

    PyObject *pyfort_result;
    int keyoption;
    long N;
    PyArrayObject* pyarray_value;
    PyObject* X;
    PyArrayObject* aX;
    int eX[1];
    PyObject* NY;
    PyArrayObject* aNY;
    int eNY[1];
    PyObject* Y;
    PyArrayObject* aY;
    int eY[1];
    long NKNOTS;
    PyArrayObject* aKNOTS;
    PyObject* rKNOTS;
    int eKNOTS[1];
    PyArrayObject* aYKNOTS;
    PyObject* rYKNOTS;
    int eYKNOTS[1];
    long LFIXED;
    double FIXED;
    double RELERROR;
    long INFO;
    aX = (PyArrayObject*) 0;
    aNY = (PyArrayObject*) 0;
    aY = (PyArrayObject*) 0;
    aKNOTS = (PyArrayObject*) 0;
    aYKNOTS = (PyArrayObject*) 0;
    keyoption = default_option;

    if(!PyArg_ParseTuple(args, "OOOld|i", &X, &NY, &Y, &LFIXED, &FIXED, &keyoption)) {
        return NULL;
    }
    if (!(aX = do_array_in ("linspline", "X", X, PyArray_DOUBLE))) goto err;
    if (!(aNY = do_array_in ("linspline", "NY", NY, PyArray_LONG))) goto err;
    if (!(aY = do_array_in ("linspline", "Y", Y, PyArray_DOUBLE))) goto err;
    N = get_fortran_dim(aX, 1);
    eX[0] = N;
    eNY[0] = N;
    eY[0] = 1;
    eKNOTS[0] = N;
    eYKNOTS[0] = N;
    if (!do_size_check ("linspline", "X", aX, 1, eX, keyoption & MIRROR_OPTION)) goto err;
    if ((keyoption & TRANSPOSE_OPTION) && (aX->nd > 1)) {
        pyarray_value = aX;
        aX = transpose_array ("linspline", "X", pyarray_value);
        Py_DECREF(pyarray_value);
        if(!aX) goto err;
    }
    pyarray_value = aX;
    aX = make_contiguous ("linspline", "X", pyarray_value);
    Py_DECREF(pyarray_value);
    if(!aX) goto err;
    if (!do_size_check ("linspline", "NY", aNY, 1, eNY, keyoption & MIRROR_OPTION)) goto err;
    if ((keyoption & TRANSPOSE_OPTION) && (aNY->nd > 1)) {
        pyarray_value = aNY;
        aNY = transpose_array ("linspline", "NY", pyarray_value);
        Py_DECREF(pyarray_value);
        if(!aNY) goto err;
    }
    pyarray_value = aNY;
    aNY = make_contiguous ("linspline", "NY", pyarray_value);
    Py_DECREF(pyarray_value);
    if(!aNY) goto err;
    if (!do_size_check ("linspline", "Y", aY, 1, eY, keyoption & MIRROR_OPTION)) goto err;
    if ((keyoption & TRANSPOSE_OPTION) && (aY->nd > 1)) {
        pyarray_value = aY;
        aY = transpose_array ("linspline", "Y", pyarray_value);
        Py_DECREF(pyarray_value);
        if(!aY) goto err;
    }
    pyarray_value = aY;
    aY = make_contiguous ("linspline", "Y", pyarray_value);
    Py_DECREF(pyarray_value);
    if(!aY) goto err;
    if (!(aKNOTS = do_array_create ("linspline", "KNOTS", PyArray_LONG, 1, eKNOTS, keyoption&MIRROR_OPTION))) goto err;
    if (!(aYKNOTS = do_array_create ("linspline", "YKNOTS", PyArray_DOUBLE, 1, eYKNOTS, keyoption&MIRROR_OPTION))) goto err;
    linspline(N, 
        (double*) (aX->data), 
        (long*) (aNY->data), 
        (double*) (aY->data), 
        &NKNOTS, 
        (long*) (aKNOTS->data), 
        (double*) (aYKNOTS->data), 
        LFIXED, 
        FIXED, 
        &RELERROR, 
        &INFO);
    rKNOTS = PyArray_Return(aKNOTS);
    rYKNOTS = PyArray_Return(aYKNOTS);
    Py_XDECREF((PyObject*) aX);
    Py_XDECREF((PyObject*) aNY);
    Py_XDECREF((PyObject*) aY);

    pyfort_result = Py_BuildValue("lOOdl",NKNOTS, rKNOTS, rYKNOTS, RELERROR, INFO);

    Py_XDECREF(rKNOTS);
    Py_XDECREF(rYKNOTS);
    return pyfort_result;
err:
    Py_XDECREF((PyObject*) aX);
    Py_XDECREF((PyObject*) aNY);
    Py_XDECREF((PyObject*) aY);
    Py_XDECREF((PyObject*) aKNOTS);
    Py_XDECREF((PyObject*) aYKNOTS);
    return NULL;
} 
/* end of wrapper for linspline */

static struct PyMethodDef linspline_methods[] = {
   {"linspline", (PyCFunction) linspline_linspline, METH_VARARGS, linspline_linspline__doc__},
   {"set_pyfort_option", (PyCFunction) set_pyfort_option, METH_VARARGS, 
           "set_pyfort_option (value) sets default value of option keyword."},
   {NULL,          NULL, 0, NULL}/* sentinel */
};

static char linspline_module_documentation[] =
"Fortran interface module linspline";

void initlinspline(void)
{
        PyObject *m, *d, *j;
 
        import_array ();
        m = Py_InitModule4("linspline", linspline_methods,
                linspline_module_documentation,
                (PyObject*)NULL,PYTHON_API_VERSION);
 
        d = PyModule_GetDict(m);
        ErrorObject = PyString_FromString("linspline.error");
        PyDict_SetItemString(d, "error", ErrorObject);
        j = PyInt_FromLong((long) TRANSPOSE_OPTION);
        PyDict_SetItemString(d, "TRANSPOSE", j);
        Py_XDECREF(j);
        j = PyInt_FromLong((long) MIRROR_OPTION);
        PyDict_SetItemString(d, "MIRROR", j);
        Py_XDECREF(j);
        j = PyInt_FromLong(0L);
        PyDict_SetItemString(d, "NONE", j);
        Py_XDECREF(j);

        if (PyErr_Occurred()) {
            Py_FatalError("can't initialize module linspline");
        }
}

/* C++ trailer */
#ifdef __CPLUSCPLUS__
}
#endif
